
module Bcrypt(		
	input wire reset,
	input wire clk,

	/*Input Handshake*/	
	input  wire  VALID_E,
	output logic READY_E,

	/*Output Handshake*/
	output logic VALID_S,
	input  wire  READY_S, 
	
	input wire unsigned [31:0]data_in,
	input wire unsigned [2:0]code,
	
	output  logic unsigned[191:0]data_out
	
);
	
	logic [31:0]sboxes[1023:0];  
	logic unsigned  [31:0]L;
	logic unsigned  [31:0]R;
	logic unsigned  [31:0]F_r;
	logic unsigned  [7:0]key_r[71:0];
	logic unsigned  [7:0]key_r2[71:0];
	logic unsigned 	[6:0]n_key_r;
	logic unsigned 	[6:0]n_key_r2;
    logic unsigned	[6:0]n_ciclic_key;
	logic unsigned  [127:0]salt_r;
	logic unsigned  [127:0]salt_r2;
	logic unsigned  [31:0]cost_r;
	logic ctrl; 
	logic unsigned [2:0][63:0]ctext;
	
	
	logic unsigned  [63:0]pApB;	
		
	logic unsigned  [31:0]parray[17:0];
	logic unsigned	[3:0] count_round;
	logic unsigned	[4:0] count_XOR_PARRAY;
	logic unsigned  [9:0] count_XOR_SBOX;
	
	logic S1S2;
	
	
	/* State machines states */
	
	enum logic[3:0] {
		IN_HANDSHAKE 					= 4'd0,
		FEISTEL							= 4'd1,
		F1 		  						= 4'd2,
		PREF							= 4'd3,
		XOR_P17_P18                		= 4'd4,
		OUT_HANDSHAKE				   	= 4'd5,
		NOP1						   	= 4'd6,
		XOR_XR_SWAP						= 4'd7,
		SWAPR							= 4'd8,
		NOP2						   	= 4'd9,
		INIT_XOR						= 4'd10,
		CONTROL_EXPAND_KEY				= 4'd11,
		INIT_PARRAY,
		CYCLIC_KEY,
		F2
	} state = IN_HANDSHAKE;
	
	enum logic[3:0] {
		XOR_PARRAY_1,
		XOR_PARRAY_2,
		XOR_SBOX_1,
		XOR_SBOX_2,
		COST_1,
		COST_2,
		COST_3,
		CRYPT_1,
		CRYPT_2,
		CRYPT_3
	} controlstate;

	always_ff@(posedge clk) begin
		if(reset==1'b1) begin
			if(state == IN_HANDSHAKE) begin//ready to read input data
				if(VALID_E ==1'b1)begin//valid data to be read
					if(code == 3'd0)begin
						cost_r <= 32'd1<<data_in;//cost
					end 
					else if(code == 3'd1)begin
						salt_r[127:96] <= data_in;
						salt_r2[127:96]	<= data_in;
					end
					else if(code == 3'd2)begin
						salt_r[95:64]   <= data_in;
						salt_r2[95:64]	<= data_in;
						n_key_r  <= 7'd0 ;
						n_key_r2 <= 7'd0;
					end
					else if(code == 3'd3)begin
						salt_r[63:32] <= data_in;
						salt_r2[63:32]	<= data_in;
					end
					else if(code == 3'd4)begin
						salt_r[31:0]   <= data_in;
						salt_r2[31:0]	<= data_in;
						n_key_r  <= 7'd0 ;
						n_key_r2 <= 7'd0;
					end
					else if(code == 3'd5)begin
						key_r[n_key_r]  <= data_in[7:0];
						key_r2[n_key_r] <= data_in[7:0];
						n_key_r <= n_key_r + 7'd1 ;
						n_key_r2 <= n_key_r2 + 7'd1;
					end 
					else if(code == 3'd6)begin
						key_r[n_key_r]  <= data_in[7:0];
						key_r2[n_key_r] <= data_in[7:0];
						n_key_r <= n_key_r + 7'd1 ;
						n_key_r2 <= n_key_r2 + 7'd1;
						
						ctrl <= 1'b0;
						S1S2 <= 1'b0;
						READY_E <= 1'b0;
						count_round <= 4'd0;
						state <= INIT_PARRAY; 
					end
				end
			end if(state == INIT_PARRAY)begin
			/*parray initialization*/
				parray[0] <= 32'h243f6a88;
				parray[1] <= 32'h85a308d3;
				parray[2] <= 32'h13198a2e;
				parray[3] <= 32'h3707344;
				parray[4] <= 32'ha4093822;
				parray[5] <= 32'h299f31d0;
				parray[6] <= 32'h82efa98;
				parray[7] <= 32'hec4e6c89;
				parray[8] <= 32'h452821e6;
				parray[9] <= 32'h38d01377;
				parray[10] <= 32'hbe5466cf;
				parray[11] <= 32'h34e90c6c;
				parray[12] <= 32'hc0ac29b7;
				parray[13] <= 32'hc97c50dd;
				parray[14] <= 32'h3f84d5b5;
				parray[15] <= 32'hb5470917;
				parray[16] <= 32'h9216d5d9;
				parray[17] <= 32'h8979fb1b;

				sboxes[0] <= 32'hd1310ba6;
				sboxes[1] <= 32'h98dfb5ac;
				sboxes[2] <= 32'h2ffd72db;
				sboxes[3] <= 32'hd01adfb7;
				sboxes[4] <= 32'hb8e1afed;
				sboxes[5] <= 32'h6a267e96;
				sboxes[6] <= 32'hba7c9045;
				sboxes[7] <= 32'hf12c7f99;
				sboxes[8] <= 32'h24a19947;
				sboxes[9] <= 32'hb3916cf7;
				sboxes[10] <= 32'h801f2e2;
				sboxes[11] <= 32'h858efc16;
				sboxes[12] <= 32'h636920d8;
				sboxes[13] <= 32'h71574e69;
				sboxes[14] <= 32'ha458fea3;
				sboxes[15] <= 32'hf4933d7e;
				sboxes[16] <= 32'hd95748f;
				sboxes[17] <= 32'h728eb658;
				sboxes[18] <= 32'h718bcd58;
				sboxes[19] <= 32'h82154aee;
				sboxes[20] <= 32'h7b54a41d;
				sboxes[21] <= 32'hc25a59b5;
				sboxes[22] <= 32'h9c30d539;
				sboxes[23] <= 32'h2af26013;
				sboxes[24] <= 32'hc5d1b023;
				sboxes[25] <= 32'h286085f0;
				sboxes[26] <= 32'hca417918;
				sboxes[27] <= 32'hb8db38ef;
				sboxes[28] <= 32'h8e79dcb0;
				sboxes[29] <= 32'h603a180e;
				sboxes[30] <= 32'h6c9e0e8b;
				sboxes[31] <= 32'hb01e8a3e;
				sboxes[32] <= 32'hd71577c1;
				sboxes[33] <= 32'hbd314b27;
				sboxes[34] <= 32'h78af2fda;
				sboxes[35] <= 32'h55605c60;
				sboxes[36] <= 32'he65525f3;
				sboxes[37] <= 32'haa55ab94;
				sboxes[38] <= 32'h57489862;
				sboxes[39] <= 32'h63e81440;
				sboxes[40] <= 32'h55ca396a;
				sboxes[41] <= 32'h2aab10b6;
				sboxes[42] <= 32'hb4cc5c34;
				sboxes[43] <= 32'h1141e8ce;
				sboxes[44] <= 32'ha15486af;
				sboxes[45] <= 32'h7c72e993;
				sboxes[46] <= 32'hb3ee1411;
				sboxes[47] <= 32'h636fbc2a;
				sboxes[48] <= 32'h2ba9c55d;
				sboxes[49] <= 32'h741831f6;
				sboxes[50] <= 32'hce5c3e16;
				sboxes[51] <= 32'h9b87931e;
				sboxes[52] <= 32'hafd6ba33;
				sboxes[53] <= 32'h6c24cf5c;
				sboxes[54] <= 32'h7a325381;
				sboxes[55] <= 32'h28958677;
				sboxes[56] <= 32'h3b8f4898;
				sboxes[57] <= 32'h6b4bb9af;
				sboxes[58] <= 32'hc4bfe81b;
				sboxes[59] <= 32'h66282193;
				sboxes[60] <= 32'h61d809cc;
				sboxes[61] <= 32'hfb21a991;
				sboxes[62] <= 32'h487cac60;
				sboxes[63] <= 32'h5dec8032;
				sboxes[64] <= 32'hef845d5d;
				sboxes[65] <= 32'he98575b1;
				sboxes[66] <= 32'hdc262302;
				sboxes[67] <= 32'heb651b88;
				sboxes[68] <= 32'h23893e81;
				sboxes[69] <= 32'hd396acc5;
				sboxes[70] <= 32'hf6d6ff3;
				sboxes[71] <= 32'h83f44239;
				sboxes[72] <= 32'h2e0b4482;
				sboxes[73] <= 32'ha4842004;
				sboxes[74] <= 32'h69c8f04a;
				sboxes[75] <= 32'h9e1f9b5e;
				sboxes[76] <= 32'h21c66842;
				sboxes[77] <= 32'hf6e96c9a;
				sboxes[78] <= 32'h670c9c61;
				sboxes[79] <= 32'habd388f0;
				sboxes[80] <= 32'h6a51a0d2;
				sboxes[81] <= 32'hd8542f68;
				sboxes[82] <= 32'h960fa728;
				sboxes[83] <= 32'hab5133a3;
				sboxes[84] <= 32'h6eef0b6c;
				sboxes[85] <= 32'h137a3be4;
				sboxes[86] <= 32'hba3bf050;
				sboxes[87] <= 32'h7efb2a98;
				sboxes[88] <= 32'ha1f1651d;
				sboxes[89] <= 32'h39af0176;
				sboxes[90] <= 32'h66ca593e;
				sboxes[91] <= 32'h82430e88;
				sboxes[92] <= 32'h8cee8619;
				sboxes[93] <= 32'h456f9fb4;
				sboxes[94] <= 32'h7d84a5c3;
				sboxes[95] <= 32'h3b8b5ebe;
				sboxes[96] <= 32'he06f75d8;
				sboxes[97] <= 32'h85c12073;
				sboxes[98] <= 32'h401a449f;
				sboxes[99] <= 32'h56c16aa6;
				sboxes[100] <= 32'h4ed3aa62;
				sboxes[101] <= 32'h363f7706;
				sboxes[102] <= 32'h1bfedf72;
				sboxes[103] <= 32'h429b023d;
				sboxes[104] <= 32'h37d0d724;
				sboxes[105] <= 32'hd00a1248;
				sboxes[106] <= 32'hdb0fead3;
				sboxes[107] <= 32'h49f1c09b;
				sboxes[108] <= 32'h75372c9;
				sboxes[109] <= 32'h80991b7b;
				sboxes[110] <= 32'h25d479d8;
				sboxes[111] <= 32'hf6e8def7;
				sboxes[112] <= 32'he3fe501a;
				sboxes[113] <= 32'hb6794c3b;
				sboxes[114] <= 32'h976ce0bd;
				sboxes[115] <= 32'h4c006ba;
				sboxes[116] <= 32'hc1a94fb6;
				sboxes[117] <= 32'h409f60c4;
				sboxes[118] <= 32'h5e5c9ec2;
				sboxes[119] <= 32'h196a2463;
				sboxes[120] <= 32'h68fb6faf;
				sboxes[121] <= 32'h3e6c53b5;
				sboxes[122] <= 32'h1339b2eb;
				sboxes[123] <= 32'h3b52ec6f;
				sboxes[124] <= 32'h6dfc511f;
				sboxes[125] <= 32'h9b30952c;
				sboxes[126] <= 32'hcc814544;
				sboxes[127] <= 32'haf5ebd09;
				sboxes[128] <= 32'hbee3d004;
				sboxes[129] <= 32'hde334afd;
				sboxes[130] <= 32'h660f2807;
				sboxes[131] <= 32'h192e4bb3;
				sboxes[132] <= 32'hc0cba857;
				sboxes[133] <= 32'h45c8740f;
				sboxes[134] <= 32'hd20b5f39;
				sboxes[135] <= 32'hb9d3fbdb;
				sboxes[136] <= 32'h5579c0bd;
				sboxes[137] <= 32'h1a60320a;
				sboxes[138] <= 32'hd6a100c6;
				sboxes[139] <= 32'h402c7279;
				sboxes[140] <= 32'h679f25fe;
				sboxes[141] <= 32'hfb1fa3cc;
				sboxes[142] <= 32'h8ea5e9f8;
				sboxes[143] <= 32'hdb3222f8;
				sboxes[144] <= 32'h3c7516df;
				sboxes[145] <= 32'hfd616b15;
				sboxes[146] <= 32'h2f501ec8;
				sboxes[147] <= 32'had0552ab;
				sboxes[148] <= 32'h323db5fa;
				sboxes[149] <= 32'hfd238760;
				sboxes[150] <= 32'h53317b48;
				sboxes[151] <= 32'h3e00df82;
				sboxes[152] <= 32'h9e5c57bb;
				sboxes[153] <= 32'hca6f8ca0;
				sboxes[154] <= 32'h1a87562e;
				sboxes[155] <= 32'hdf1769db;
				sboxes[156] <= 32'hd542a8f6;
				sboxes[157] <= 32'h287effc3;
				sboxes[158] <= 32'hac6732c6;
				sboxes[159] <= 32'h8c4f5573;
				sboxes[160] <= 32'h695b27b0;
				sboxes[161] <= 32'hbbca58c8;
				sboxes[162] <= 32'he1ffa35d;
				sboxes[163] <= 32'hb8f011a0;
				sboxes[164] <= 32'h10fa3d98;
				sboxes[165] <= 32'hfd2183b8;
				sboxes[166] <= 32'h4afcb56c;
				sboxes[167] <= 32'h2dd1d35b;
				sboxes[168] <= 32'h9a53e479;
				sboxes[169] <= 32'hb6f84565;
				sboxes[170] <= 32'hd28e49bc;
				sboxes[171] <= 32'h4bfb9790;
				sboxes[172] <= 32'he1ddf2da;
				sboxes[173] <= 32'ha4cb7e33;
				sboxes[174] <= 32'h62fb1341;
				sboxes[175] <= 32'hcee4c6e8;
				sboxes[176] <= 32'hef20cada;
				sboxes[177] <= 32'h36774c01;
				sboxes[178] <= 32'hd07e9efe;
				sboxes[179] <= 32'h2bf11fb4;
				sboxes[180] <= 32'h95dbda4d;
				sboxes[181] <= 32'hae909198;
				sboxes[182] <= 32'heaad8e71;
				sboxes[183] <= 32'h6b93d5a0;
				sboxes[184] <= 32'hd08ed1d0;
				sboxes[185] <= 32'hafc725e0;
				sboxes[186] <= 32'h8e3c5b2f;
				sboxes[187] <= 32'h8e7594b7;
				sboxes[188] <= 32'h8ff6e2fb;
				sboxes[189] <= 32'hf2122b64;
				sboxes[190] <= 32'h8888b812;
				sboxes[191] <= 32'h900df01c;
				sboxes[192] <= 32'h4fad5ea0;
				sboxes[193] <= 32'h688fc31c;
				sboxes[194] <= 32'hd1cff191;
				sboxes[195] <= 32'hb3a8c1ad;
				sboxes[196] <= 32'h2f2f2218;
				sboxes[197] <= 32'hbe0e1777;
				sboxes[198] <= 32'hea752dfe;
				sboxes[199] <= 32'h8b021fa1;
				sboxes[200] <= 32'he5a0cc0f;
				sboxes[201] <= 32'hb56f74e8;
				sboxes[202] <= 32'h18acf3d6;
				sboxes[203] <= 32'hce89e299;
				sboxes[204] <= 32'hb4a84fe0;
				sboxes[205] <= 32'hfd13e0b7;
				sboxes[206] <= 32'h7cc43b81;
				sboxes[207] <= 32'hd2ada8d9;
				sboxes[208] <= 32'h165fa266;
				sboxes[209] <= 32'h80957705;
				sboxes[210] <= 32'h93cc7314;
				sboxes[211] <= 32'h211a1477;
				sboxes[212] <= 32'he6ad2065;
				sboxes[213] <= 32'h77b5fa86;
				sboxes[214] <= 32'hc75442f5;
				sboxes[215] <= 32'hfb9d35cf;
				sboxes[216] <= 32'hebcdaf0c;
				sboxes[217] <= 32'h7b3e89a0;
				sboxes[218] <= 32'hd6411bd3;
				sboxes[219] <= 32'hae1e7e49;
				sboxes[220] <= 32'h250e2d;
				sboxes[221] <= 32'h2071b35e;
				sboxes[222] <= 32'h226800bb;
				sboxes[223] <= 32'h57b8e0af;
				sboxes[224] <= 32'h2464369b;
				sboxes[225] <= 32'hf009b91e;
				sboxes[226] <= 32'h5563911d;
				sboxes[227] <= 32'h59dfa6aa;
				sboxes[228] <= 32'h78c14389;
				sboxes[229] <= 32'hd95a537f;
				sboxes[230] <= 32'h207d5ba2;
				sboxes[231] <= 32'h2e5b9c5;
				sboxes[232] <= 32'h83260376;
				sboxes[233] <= 32'h6295cfa9;
				sboxes[234] <= 32'h11c81968;
				sboxes[235] <= 32'h4e734a41;
				sboxes[236] <= 32'hb3472dca;
				sboxes[237] <= 32'h7b14a94a;
				sboxes[238] <= 32'h1b510052;
				sboxes[239] <= 32'h9a532915;
				sboxes[240] <= 32'hd60f573f;
				sboxes[241] <= 32'hbc9bc6e4;
				sboxes[242] <= 32'h2b60a476;
				sboxes[243] <= 32'h81e67400;
				sboxes[244] <= 32'h8ba6fb5;
				sboxes[245] <= 32'h571be91f;
				sboxes[246] <= 32'hf296ec6b;
				sboxes[247] <= 32'h2a0dd915;
				sboxes[248] <= 32'hb6636521;
				sboxes[249] <= 32'he7b9f9b6;
				sboxes[250] <= 32'hff34052e;
				sboxes[251] <= 32'hc5855664;
				sboxes[252] <= 32'h53b02d5d;
				sboxes[253] <= 32'ha99f8fa1;
				sboxes[254] <= 32'h8ba4799;
				sboxes[255] <= 32'h6e85076a;
				sboxes[256] <= 32'h4b7a70e9;
				sboxes[257] <= 32'hb5b32944;
				sboxes[258] <= 32'hdb75092e;
				sboxes[259] <= 32'hc4192623;
				sboxes[260] <= 32'had6ea6b0;
				sboxes[261] <= 32'h49a7df7d;
				sboxes[262] <= 32'h9cee60b8;
				sboxes[263] <= 32'h8fedb266;
				sboxes[264] <= 32'hecaa8c71;
				sboxes[265] <= 32'h699a17ff;
				sboxes[266] <= 32'h5664526c;
				sboxes[267] <= 32'hc2b19ee1;
				sboxes[268] <= 32'h193602a5;
				sboxes[269] <= 32'h75094c29;
				sboxes[270] <= 32'ha0591340;
				sboxes[271] <= 32'he4183a3e;
				sboxes[272] <= 32'h3f54989a;
				sboxes[273] <= 32'h5b429d65;
				sboxes[274] <= 32'h6b8fe4d6;
				sboxes[275] <= 32'h99f73fd6;
				sboxes[276] <= 32'ha1d29c07;
				sboxes[277] <= 32'hefe830f5;
				sboxes[278] <= 32'h4d2d38e6;
				sboxes[279] <= 32'hf0255dc1;
				sboxes[280] <= 32'h4cdd2086;
				sboxes[281] <= 32'h8470eb26;
				sboxes[282] <= 32'h6382e9c6;
				sboxes[283] <= 32'h21ecc5e;
				sboxes[284] <= 32'h9686b3f;
				sboxes[285] <= 32'h3ebaefc9;
				sboxes[286] <= 32'h3c971814;
				sboxes[287] <= 32'h6b6a70a1;
				sboxes[288] <= 32'h687f3584;
				sboxes[289] <= 32'h52a0e286;
				sboxes[290] <= 32'hb79c5305;
				sboxes[291] <= 32'haa500737;
				sboxes[292] <= 32'h3e07841c;
				sboxes[293] <= 32'h7fdeae5c;
				sboxes[294] <= 32'h8e7d44ec;
				sboxes[295] <= 32'h5716f2b8;
				sboxes[296] <= 32'hb03ada37;
				sboxes[297] <= 32'hf0500c0d;
				sboxes[298] <= 32'hf01c1f04;
				sboxes[299] <= 32'h200b3ff;
				sboxes[300] <= 32'hae0cf51a;
				sboxes[301] <= 32'h3cb574b2;
				sboxes[302] <= 32'h25837a58;
				sboxes[303] <= 32'hdc0921bd;
				sboxes[304] <= 32'hd19113f9;
				sboxes[305] <= 32'h7ca92ff6;
				sboxes[306] <= 32'h94324773;
				sboxes[307] <= 32'h22f54701;
				sboxes[308] <= 32'h3ae5e581;
				sboxes[309] <= 32'h37c2dadc;
				sboxes[310] <= 32'hc8b57634;
				sboxes[311] <= 32'h9af3dda7;
				sboxes[312] <= 32'ha9446146;
				sboxes[313] <= 32'hfd0030e;
				sboxes[314] <= 32'hecc8c73e;
				sboxes[315] <= 32'ha4751e41;
				sboxes[316] <= 32'he238cd99;
				sboxes[317] <= 32'h3bea0e2f;
				sboxes[318] <= 32'h3280bba1;
				sboxes[319] <= 32'h183eb331;
				sboxes[320] <= 32'h4e548b38;
				sboxes[321] <= 32'h4f6db908;
				sboxes[322] <= 32'h6f420d03;
				sboxes[323] <= 32'hf60a04bf;
				sboxes[324] <= 32'h2cb81290;
				sboxes[325] <= 32'h24977c79;
				sboxes[326] <= 32'h5679b072;
				sboxes[327] <= 32'hbcaf89af;
				sboxes[328] <= 32'hde9a771f;
				sboxes[329] <= 32'hd9930810;
				sboxes[330] <= 32'hb38bae12;
				sboxes[331] <= 32'hdccf3f2e;
				sboxes[332] <= 32'h5512721f;
				sboxes[333] <= 32'h2e6b7124;
				sboxes[334] <= 32'h501adde6;
				sboxes[335] <= 32'h9f84cd87;
				sboxes[336] <= 32'h7a584718;
				sboxes[337] <= 32'h7408da17;
				sboxes[338] <= 32'hbc9f9abc;
				sboxes[339] <= 32'he94b7d8c;
				sboxes[340] <= 32'hec7aec3a;
				sboxes[341] <= 32'hdb851dfa;
				sboxes[342] <= 32'h63094366;
				sboxes[343] <= 32'hc464c3d2;
				sboxes[344] <= 32'hef1c1847;
				sboxes[345] <= 32'h3215d908;
				sboxes[346] <= 32'hdd433b37;
				sboxes[347] <= 32'h24c2ba16;
				sboxes[348] <= 32'h12a14d43;
				sboxes[349] <= 32'h2a65c451;
				sboxes[350] <= 32'h50940002;
				sboxes[351] <= 32'h133ae4dd;
				sboxes[352] <= 32'h71dff89e;
				sboxes[353] <= 32'h10314e55;
				sboxes[354] <= 32'h81ac77d6;
				sboxes[355] <= 32'h5f11199b;
				sboxes[356] <= 32'h43556f1;
				sboxes[357] <= 32'hd7a3c76b;
				sboxes[358] <= 32'h3c11183b;
				sboxes[359] <= 32'h5924a509;
				sboxes[360] <= 32'hf28fe6ed;
				sboxes[361] <= 32'h97f1fbfa;
				sboxes[362] <= 32'h9ebabf2c;
				sboxes[363] <= 32'h1e153c6e;
				sboxes[364] <= 32'h86e34570;
				sboxes[365] <= 32'heae96fb1;
				sboxes[366] <= 32'h860e5e0a;
				sboxes[367] <= 32'h5a3e2ab3;
				sboxes[368] <= 32'h771fe71c;
				sboxes[369] <= 32'h4e3d06fa;
				sboxes[370] <= 32'h2965dcb9;
				sboxes[371] <= 32'h99e71d0f;
				sboxes[372] <= 32'h803e89d6;
				sboxes[373] <= 32'h5266c825;
				sboxes[374] <= 32'h2e4cc978;
				sboxes[375] <= 32'h9c10b36a;
				sboxes[376] <= 32'hc6150eba;
				sboxes[377] <= 32'h94e2ea78;
				sboxes[378] <= 32'ha5fc3c53;
				sboxes[379] <= 32'h1e0a2df4;
				sboxes[380] <= 32'hf2f74ea7;
				sboxes[381] <= 32'h361d2b3d;
				sboxes[382] <= 32'h1939260f;
				sboxes[383] <= 32'h19c27960;
				sboxes[384] <= 32'h5223a708;
				sboxes[385] <= 32'hf71312b6;
				sboxes[386] <= 32'hebadfe6e;
				sboxes[387] <= 32'heac31f66;
				sboxes[388] <= 32'he3bc4595;
				sboxes[389] <= 32'ha67bc883;
				sboxes[390] <= 32'hb17f37d1;
				sboxes[391] <= 32'h18cff28;
				sboxes[392] <= 32'hc332ddef;
				sboxes[393] <= 32'hbe6c5aa5;
				sboxes[394] <= 32'h65582185;
				sboxes[395] <= 32'h68ab9802;
				sboxes[396] <= 32'heecea50f;
				sboxes[397] <= 32'hdb2f953b;
				sboxes[398] <= 32'h2aef7dad;
				sboxes[399] <= 32'h5b6e2f84;
				sboxes[400] <= 32'h1521b628;
				sboxes[401] <= 32'h29076170;
				sboxes[402] <= 32'hecdd4775;
				sboxes[403] <= 32'h619f1510;
				sboxes[404] <= 32'h13cca830;
				sboxes[405] <= 32'heb61bd96;
				sboxes[406] <= 32'h334fe1e;
				sboxes[407] <= 32'haa0363cf;
				sboxes[408] <= 32'hb5735c90;
				sboxes[409] <= 32'h4c70a239;
				sboxes[410] <= 32'hd59e9e0b;
				sboxes[411] <= 32'hcbaade14;
				sboxes[412] <= 32'heecc86bc;
				sboxes[413] <= 32'h60622ca7;
				sboxes[414] <= 32'h9cab5cab;
				sboxes[415] <= 32'hb2f3846e;
				sboxes[416] <= 32'h648b1eaf;
				sboxes[417] <= 32'h19bdf0ca;
				sboxes[418] <= 32'ha02369b9;
				sboxes[419] <= 32'h655abb50;
				sboxes[420] <= 32'h40685a32;
				sboxes[421] <= 32'h3c2ab4b3;
				sboxes[422] <= 32'h319ee9d5;
				sboxes[423] <= 32'hc021b8f7;
				sboxes[424] <= 32'h9b540b19;
				sboxes[425] <= 32'h875fa099;
				sboxes[426] <= 32'h95f7997e;
				sboxes[427] <= 32'h623d7da8;
				sboxes[428] <= 32'hf837889a;
				sboxes[429] <= 32'h97e32d77;
				sboxes[430] <= 32'h11ed935f;
				sboxes[431] <= 32'h16681281;
				sboxes[432] <= 32'he358829;
				sboxes[433] <= 32'hc7e61fd6;
				sboxes[434] <= 32'h96dedfa1;
				sboxes[435] <= 32'h7858ba99;
				sboxes[436] <= 32'h57f584a5;
				sboxes[437] <= 32'h1b227263;
				sboxes[438] <= 32'h9b83c3ff;
				sboxes[439] <= 32'h1ac24696;
				sboxes[440] <= 32'hcdb30aeb;
				sboxes[441] <= 32'h532e3054;
				sboxes[442] <= 32'h8fd948e4;
				sboxes[443] <= 32'h6dbc3128;
				sboxes[444] <= 32'h58ebf2ef;
				sboxes[445] <= 32'h34c6ffea;
				sboxes[446] <= 32'hfe28ed61;
				sboxes[447] <= 32'hee7c3c73;
				sboxes[448] <= 32'h5d4a14d9;
				sboxes[449] <= 32'he864b7e3;
				sboxes[450] <= 32'h42105d14;
				sboxes[451] <= 32'h203e13e0;
				sboxes[452] <= 32'h45eee2b6;
				sboxes[453] <= 32'ha3aaabea;
				sboxes[454] <= 32'hdb6c4f15;
				sboxes[455] <= 32'hfacb4fd0;
				sboxes[456] <= 32'hc742f442;
				sboxes[457] <= 32'hef6abbb5;
				sboxes[458] <= 32'h654f3b1d;
				sboxes[459] <= 32'h41cd2105;
				sboxes[460] <= 32'hd81e799e;
				sboxes[461] <= 32'h86854dc7;
				sboxes[462] <= 32'he44b476a;
				sboxes[463] <= 32'h3d816250;
				sboxes[464] <= 32'hcf62a1f2;
				sboxes[465] <= 32'h5b8d2646;
				sboxes[466] <= 32'hfc8883a0;
				sboxes[467] <= 32'hc1c7b6a3;
				sboxes[468] <= 32'h7f1524c3;
				sboxes[469] <= 32'h69cb7492;
				sboxes[470] <= 32'h47848a0b;
				sboxes[471] <= 32'h5692b285;
				sboxes[472] <= 32'h95bbf00;
				sboxes[473] <= 32'had19489d;
				sboxes[474] <= 32'h1462b174;
				sboxes[475] <= 32'h23820e00;
				sboxes[476] <= 32'h58428d2a;
				sboxes[477] <= 32'hc55f5ea;
				sboxes[478] <= 32'h1dadf43e;
				sboxes[479] <= 32'h233f7061;
				sboxes[480] <= 32'h3372f092;
				sboxes[481] <= 32'h8d937e41;
				sboxes[482] <= 32'hd65fecf1;
				sboxes[483] <= 32'h6c223bdb;
				sboxes[484] <= 32'h7cde3759;
				sboxes[485] <= 32'hcbee7460;
				sboxes[486] <= 32'h4085f2a7;
				sboxes[487] <= 32'hce77326e;
				sboxes[488] <= 32'ha6078084;
				sboxes[489] <= 32'h19f8509e;
				sboxes[490] <= 32'he8efd855;
				sboxes[491] <= 32'h61d99735;
				sboxes[492] <= 32'ha969a7aa;
				sboxes[493] <= 32'hc50c06c2;
				sboxes[494] <= 32'h5a04abfc;
				sboxes[495] <= 32'h800bcadc;
				sboxes[496] <= 32'h9e447a2e;
				sboxes[497] <= 32'hc3453484;
				sboxes[498] <= 32'hfdd56705;
				sboxes[499] <= 32'he1e9ec9;
				sboxes[500] <= 32'hdb73dbd3;
				sboxes[501] <= 32'h105588cd;
				sboxes[502] <= 32'h675fda79;
				sboxes[503] <= 32'he3674340;
				sboxes[504] <= 32'hc5c43465;
				sboxes[505] <= 32'h713e38d8;
				sboxes[506] <= 32'h3d28f89e;
				sboxes[507] <= 32'hf16dff20;
				sboxes[508] <= 32'h153e21e7;
				sboxes[509] <= 32'h8fb03d4a;
				sboxes[510] <= 32'he6e39f2b;
				sboxes[511] <= 32'hdb83adf7;
				sboxes[512] <= 32'he93d5a68;
				sboxes[513] <= 32'h948140f7;
				sboxes[514] <= 32'hf64c261c;
				sboxes[515] <= 32'h94692934;
				sboxes[516] <= 32'h411520f7;
				sboxes[517] <= 32'h7602d4f7;
				sboxes[518] <= 32'hbcf46b2e;
				sboxes[519] <= 32'hd4a20068;
				sboxes[520] <= 32'hd4082471;
				sboxes[521] <= 32'h3320f46a;
				sboxes[522] <= 32'h43b7d4b7;
				sboxes[523] <= 32'h500061af;
				sboxes[524] <= 32'h1e39f62e;
				sboxes[525] <= 32'h97244546;
				sboxes[526] <= 32'h14214f74;
				sboxes[527] <= 32'hbf8b8840;
				sboxes[528] <= 32'h4d95fc1d;
				sboxes[529] <= 32'h96b591af;
				sboxes[530] <= 32'h70f4ddd3;
				sboxes[531] <= 32'h66a02f45;
				sboxes[532] <= 32'hbfbc09ec;
				sboxes[533] <= 32'h3bd9785;
				sboxes[534] <= 32'h7fac6dd0;
				sboxes[535] <= 32'h31cb8504;
				sboxes[536] <= 32'h96eb27b3;
				sboxes[537] <= 32'h55fd3941;
				sboxes[538] <= 32'hda2547e6;
				sboxes[539] <= 32'habca0a9a;
				sboxes[540] <= 32'h28507825;
				sboxes[541] <= 32'h530429f4;
				sboxes[542] <= 32'ha2c86da;
				sboxes[543] <= 32'he9b66dfb;
				sboxes[544] <= 32'h68dc1462;
				sboxes[545] <= 32'hd7486900;
				sboxes[546] <= 32'h680ec0a4;
				sboxes[547] <= 32'h27a18dee;
				sboxes[548] <= 32'h4f3ffea2;
				sboxes[549] <= 32'he887ad8c;
				sboxes[550] <= 32'hb58ce006;
				sboxes[551] <= 32'h7af4d6b6;
				sboxes[552] <= 32'haace1e7c;
				sboxes[553] <= 32'hd3375fec;
				sboxes[554] <= 32'hce78a399;
				sboxes[555] <= 32'h406b2a42;
				sboxes[556] <= 32'h20fe9e35;
				sboxes[557] <= 32'hd9f385b9;
				sboxes[558] <= 32'hee39d7ab;
				sboxes[559] <= 32'h3b124e8b;
				sboxes[560] <= 32'h1dc9faf7;
				sboxes[561] <= 32'h4b6d1856;
				sboxes[562] <= 32'h26a36631;
				sboxes[563] <= 32'heae397b2;
				sboxes[564] <= 32'h3a6efa74;
				sboxes[565] <= 32'hdd5b4332;
				sboxes[566] <= 32'h6841e7f7;
				sboxes[567] <= 32'hca7820fb;
				sboxes[568] <= 32'hfb0af54e;
				sboxes[569] <= 32'hd8feb397;
				sboxes[570] <= 32'h454056ac;
				sboxes[571] <= 32'hba489527;
				sboxes[572] <= 32'h55533a3a;
				sboxes[573] <= 32'h20838d87;
				sboxes[574] <= 32'hfe6ba9b7;
				sboxes[575] <= 32'hd096954b;
				sboxes[576] <= 32'h55a867bc;
				sboxes[577] <= 32'ha1159a58;
				sboxes[578] <= 32'hcca92963;
				sboxes[579] <= 32'h99e1db33;
				sboxes[580] <= 32'ha62a4a56;
				sboxes[581] <= 32'h3f3125f9;
				sboxes[582] <= 32'h5ef47e1c;
				sboxes[583] <= 32'h9029317c;
				sboxes[584] <= 32'hfdf8e802;
				sboxes[585] <= 32'h4272f70;
				sboxes[586] <= 32'h80bb155c;
				sboxes[587] <= 32'h5282ce3;
				sboxes[588] <= 32'h95c11548;
				sboxes[589] <= 32'he4c66d22;
				sboxes[590] <= 32'h48c1133f;
				sboxes[591] <= 32'hc70f86dc;
				sboxes[592] <= 32'h7f9c9ee;
				sboxes[593] <= 32'h41041f0f;
				sboxes[594] <= 32'h404779a4;
				sboxes[595] <= 32'h5d886e17;
				sboxes[596] <= 32'h325f51eb;
				sboxes[597] <= 32'hd59bc0d1;
				sboxes[598] <= 32'hf2bcc18f;
				sboxes[599] <= 32'h41113564;
				sboxes[600] <= 32'h257b7834;
				sboxes[601] <= 32'h602a9c60;
				sboxes[602] <= 32'hdff8e8a3;
				sboxes[603] <= 32'h1f636c1b;
				sboxes[604] <= 32'he12b4c2;
				sboxes[605] <= 32'h2e1329e;
				sboxes[606] <= 32'haf664fd1;
				sboxes[607] <= 32'hcad18115;
				sboxes[608] <= 32'h6b2395e0;
				sboxes[609] <= 32'h333e92e1;
				sboxes[610] <= 32'h3b240b62;
				sboxes[611] <= 32'heebeb922;
				sboxes[612] <= 32'h85b2a20e;
				sboxes[613] <= 32'he6ba0d99;
				sboxes[614] <= 32'hde720c8c;
				sboxes[615] <= 32'h2da2f728;
				sboxes[616] <= 32'hd0127845;
				sboxes[617] <= 32'h95b794fd;
				sboxes[618] <= 32'h647d0862;
				sboxes[619] <= 32'he7ccf5f0;
				sboxes[620] <= 32'h5449a36f;
				sboxes[621] <= 32'h877d48fa;
				sboxes[622] <= 32'hc39dfd27;
				sboxes[623] <= 32'hf33e8d1e;
				sboxes[624] <= 32'ha476341;
				sboxes[625] <= 32'h992eff74;
				sboxes[626] <= 32'h3a6f6eab;
				sboxes[627] <= 32'hf4f8fd37;
				sboxes[628] <= 32'ha812dc60;
				sboxes[629] <= 32'ha1ebddf8;
				sboxes[630] <= 32'h991be14c;
				sboxes[631] <= 32'hdb6e6b0d;
				sboxes[632] <= 32'hc67b5510;
				sboxes[633] <= 32'h6d672c37;
				sboxes[634] <= 32'h2765d43b;
				sboxes[635] <= 32'hdcd0e804;
				sboxes[636] <= 32'hf1290dc7;
				sboxes[637] <= 32'hcc00ffa3;
				sboxes[638] <= 32'hb5390f92;
				sboxes[639] <= 32'h690fed0b;
				sboxes[640] <= 32'h667b9ffb;
				sboxes[641] <= 32'hcedb7d9c;
				sboxes[642] <= 32'ha091cf0b;
				sboxes[643] <= 32'hd9155ea3;
				sboxes[644] <= 32'hbb132f88;
				sboxes[645] <= 32'h515bad24;
				sboxes[646] <= 32'h7b9479bf;
				sboxes[647] <= 32'h763bd6eb;
				sboxes[648] <= 32'h37392eb3;
				sboxes[649] <= 32'hcc115979;
				sboxes[650] <= 32'h8026e297;
				sboxes[651] <= 32'hf42e312d;
				sboxes[652] <= 32'h6842ada7;
				sboxes[653] <= 32'hc66a2b3b;
				sboxes[654] <= 32'h12754ccc;
				sboxes[655] <= 32'h782ef11c;
				sboxes[656] <= 32'h6a124237;
				sboxes[657] <= 32'hb79251e7;
				sboxes[658] <= 32'h6a1bbe6;
				sboxes[659] <= 32'h4bfb6350;
				sboxes[660] <= 32'h1a6b1018;
				sboxes[661] <= 32'h11caedfa;
				sboxes[662] <= 32'h3d25bdd8;
				sboxes[663] <= 32'he2e1c3c9;
				sboxes[664] <= 32'h44421659;
				sboxes[665] <= 32'ha121386;
				sboxes[666] <= 32'hd90cec6e;
				sboxes[667] <= 32'hd5abea2a;
				sboxes[668] <= 32'h64af674e;
				sboxes[669] <= 32'hda86a85f;
				sboxes[670] <= 32'hbebfe988;
				sboxes[671] <= 32'h64e4c3fe;
				sboxes[672] <= 32'h9dbc8057;
				sboxes[673] <= 32'hf0f7c086;
				sboxes[674] <= 32'h60787bf8;
				sboxes[675] <= 32'h6003604d;
				sboxes[676] <= 32'hd1fd8346;
				sboxes[677] <= 32'hf6381fb0;
				sboxes[678] <= 32'h7745ae04;
				sboxes[679] <= 32'hd736fccc;
				sboxes[680] <= 32'h83426b33;
				sboxes[681] <= 32'hf01eab71;
				sboxes[682] <= 32'hb0804187;
				sboxes[683] <= 32'h3c005e5f;
				sboxes[684] <= 32'h77a057be;
				sboxes[685] <= 32'hbde8ae24;
				sboxes[686] <= 32'h55464299;
				sboxes[687] <= 32'hbf582e61;
				sboxes[688] <= 32'h4e58f48f;
				sboxes[689] <= 32'hf2ddfda2;
				sboxes[690] <= 32'hf474ef38;
				sboxes[691] <= 32'h8789bdc2;
				sboxes[692] <= 32'h5366f9c3;
				sboxes[693] <= 32'hc8b38e74;
				sboxes[694] <= 32'hb475f255;
				sboxes[695] <= 32'h46fcd9b9;
				sboxes[696] <= 32'h7aeb2661;
				sboxes[697] <= 32'h8b1ddf84;
				sboxes[698] <= 32'h846a0e79;
				sboxes[699] <= 32'h915f95e2;
				sboxes[700] <= 32'h466e598e;
				sboxes[701] <= 32'h20b45770;
				sboxes[702] <= 32'h8cd55591;
				sboxes[703] <= 32'hc902de4c;
				sboxes[704] <= 32'hb90bace1;
				sboxes[705] <= 32'hbb8205d0;
				sboxes[706] <= 32'h11a86248;
				sboxes[707] <= 32'h7574a99e;
				sboxes[708] <= 32'hb77f19b6;
				sboxes[709] <= 32'he0a9dc09;
				sboxes[710] <= 32'h662d09a1;
				sboxes[711] <= 32'hc4324633;
				sboxes[712] <= 32'he85a1f02;
				sboxes[713] <= 32'h9f0be8c;
				sboxes[714] <= 32'h4a99a025;
				sboxes[715] <= 32'h1d6efe10;
				sboxes[716] <= 32'h1ab93d1d;
				sboxes[717] <= 32'hba5a4df;
				sboxes[718] <= 32'ha186f20f;
				sboxes[719] <= 32'h2868f169;
				sboxes[720] <= 32'hdcb7da83;
				sboxes[721] <= 32'h573906fe;
				sboxes[722] <= 32'ha1e2ce9b;
				sboxes[723] <= 32'h4fcd7f52;
				sboxes[724] <= 32'h50115e01;
				sboxes[725] <= 32'ha70683fa;
				sboxes[726] <= 32'ha002b5c4;
				sboxes[727] <= 32'hde6d027;
				sboxes[728] <= 32'h9af88c27;
				sboxes[729] <= 32'h773f8641;
				sboxes[730] <= 32'hc3604c06;
				sboxes[731] <= 32'h61a806b5;
				sboxes[732] <= 32'hf0177a28;
				sboxes[733] <= 32'hc0f586e0;
				sboxes[734] <= 32'h6058aa;
				sboxes[735] <= 32'h30dc7d62;
				sboxes[736] <= 32'h11e69ed7;
				sboxes[737] <= 32'h2338ea63;
				sboxes[738] <= 32'h53c2dd94;
				sboxes[739] <= 32'hc2c21634;
				sboxes[740] <= 32'hbbcbee56;
				sboxes[741] <= 32'h90bcb6de;
				sboxes[742] <= 32'hebfc7da1;
				sboxes[743] <= 32'hce591d76;
				sboxes[744] <= 32'h6f05e409;
				sboxes[745] <= 32'h4b7c0188;
				sboxes[746] <= 32'h39720a3d;
				sboxes[747] <= 32'h7c927c24;
				sboxes[748] <= 32'h86e3725f;
				sboxes[749] <= 32'h724d9db9;
				sboxes[750] <= 32'h1ac15bb4;
				sboxes[751] <= 32'hd39eb8fc;
				sboxes[752] <= 32'hed545578;
				sboxes[753] <= 32'h8fca5b5;
				sboxes[754] <= 32'hd83d7cd3;
				sboxes[755] <= 32'h4dad0fc4;
				sboxes[756] <= 32'h1e50ef5e;
				sboxes[757] <= 32'hb161e6f8;
				sboxes[758] <= 32'ha28514d9;
				sboxes[759] <= 32'h6c51133c;
				sboxes[760] <= 32'h6fd5c7e7;
				sboxes[761] <= 32'h56e14ec4;
				sboxes[762] <= 32'h362abfce;
				sboxes[763] <= 32'hddc6c837;
				sboxes[764] <= 32'hd79a3234;
				sboxes[765] <= 32'h92638212;
				sboxes[766] <= 32'h670efa8e;
				sboxes[767] <= 32'h406000e0;
				sboxes[768] <= 32'h3a39ce37;
				sboxes[769] <= 32'hd3faf5cf;
				sboxes[770] <= 32'habc27737;
				sboxes[771] <= 32'h5ac52d1b;
				sboxes[772] <= 32'h5cb0679e;
				sboxes[773] <= 32'h4fa33742;
				sboxes[774] <= 32'hd3822740;
				sboxes[775] <= 32'h99bc9bbe;
				sboxes[776] <= 32'hd5118e9d;
				sboxes[777] <= 32'hbf0f7315;
				sboxes[778] <= 32'hd62d1c7e;
				sboxes[779] <= 32'hc700c47b;
				sboxes[780] <= 32'hb78c1b6b;
				sboxes[781] <= 32'h21a19045;
				sboxes[782] <= 32'hb26eb1be;
				sboxes[783] <= 32'h6a366eb4;
				sboxes[784] <= 32'h5748ab2f;
				sboxes[785] <= 32'hbc946e79;
				sboxes[786] <= 32'hc6a376d2;
				sboxes[787] <= 32'h6549c2c8;
				sboxes[788] <= 32'h530ff8ee;
				sboxes[789] <= 32'h468dde7d;
				sboxes[790] <= 32'hd5730a1d;
				sboxes[791] <= 32'h4cd04dc6;
				sboxes[792] <= 32'h2939bbdb;
				sboxes[793] <= 32'ha9ba4650;
				sboxes[794] <= 32'hac9526e8;
				sboxes[795] <= 32'hbe5ee304;
				sboxes[796] <= 32'ha1fad5f0;
				sboxes[797] <= 32'h6a2d519a;
				sboxes[798] <= 32'h63ef8ce2;
				sboxes[799] <= 32'h9a86ee22;
				sboxes[800] <= 32'hc089c2b8;
				sboxes[801] <= 32'h43242ef6;
				sboxes[802] <= 32'ha51e03aa;
				sboxes[803] <= 32'h9cf2d0a4;
				sboxes[804] <= 32'h83c061ba;
				sboxes[805] <= 32'h9be96a4d;
				sboxes[806] <= 32'h8fe51550;
				sboxes[807] <= 32'hba645bd6;
				sboxes[808] <= 32'h2826a2f9;
				sboxes[809] <= 32'ha73a3ae1;
				sboxes[810] <= 32'h4ba99586;
				sboxes[811] <= 32'hef5562e9;
				sboxes[812] <= 32'hc72fefd3;
				sboxes[813] <= 32'hf752f7da;
				sboxes[814] <= 32'h3f046f69;
				sboxes[815] <= 32'h77fa0a59;
				sboxes[816] <= 32'h80e4a915;
				sboxes[817] <= 32'h87b08601;
				sboxes[818] <= 32'h9b09e6ad;
				sboxes[819] <= 32'h3b3ee593;
				sboxes[820] <= 32'he990fd5a;
				sboxes[821] <= 32'h9e34d797;
				sboxes[822] <= 32'h2cf0b7d9;
				sboxes[823] <= 32'h22b8b51;
				sboxes[824] <= 32'h96d5ac3a;
				sboxes[825] <= 32'h17da67d;
				sboxes[826] <= 32'hd1cf3ed6;
				sboxes[827] <= 32'h7c7d2d28;
				sboxes[828] <= 32'h1f9f25cf;
				sboxes[829] <= 32'hadf2b89b;
				sboxes[830] <= 32'h5ad6b472;
				sboxes[831] <= 32'h5a88f54c;
				sboxes[832] <= 32'he029ac71;
				sboxes[833] <= 32'he019a5e6;
				sboxes[834] <= 32'h47b0acfd;
				sboxes[835] <= 32'hed93fa9b;
				sboxes[836] <= 32'he8d3c48d;
				sboxes[837] <= 32'h283b57cc;
				sboxes[838] <= 32'hf8d56629;
				sboxes[839] <= 32'h79132e28;
				sboxes[840] <= 32'h785f0191;
				sboxes[841] <= 32'hed756055;
				sboxes[842] <= 32'hf7960e44;
				sboxes[843] <= 32'he3d35e8c;
				sboxes[844] <= 32'h15056dd4;
				sboxes[845] <= 32'h88f46dba;
				sboxes[846] <= 32'h3a16125;
				sboxes[847] <= 32'h564f0bd;
				sboxes[848] <= 32'hc3eb9e15;
				sboxes[849] <= 32'h3c9057a2;
				sboxes[850] <= 32'h97271aec;
				sboxes[851] <= 32'ha93a072a;
				sboxes[852] <= 32'h1b3f6d9b;
				sboxes[853] <= 32'h1e6321f5;
				sboxes[854] <= 32'hf59c66fb;
				sboxes[855] <= 32'h26dcf319;
				sboxes[856] <= 32'h7533d928;
				sboxes[857] <= 32'hb155fdf5;
				sboxes[858] <= 32'h3563482;
				sboxes[859] <= 32'h8aba3cbb;
				sboxes[860] <= 32'h28517711;
				sboxes[861] <= 32'hc20ad9f8;
				sboxes[862] <= 32'habcc5167;
				sboxes[863] <= 32'hccad925f;
				sboxes[864] <= 32'h4de81751;
				sboxes[865] <= 32'h3830dc8e;
				sboxes[866] <= 32'h379d5862;
				sboxes[867] <= 32'h9320f991;
				sboxes[868] <= 32'hea7a90c2;
				sboxes[869] <= 32'hfb3e7bce;
				sboxes[870] <= 32'h5121ce64;
				sboxes[871] <= 32'h774fbe32;
				sboxes[872] <= 32'ha8b6e37e;
				sboxes[873] <= 32'hc3293d46;
				sboxes[874] <= 32'h48de5369;
				sboxes[875] <= 32'h6413e680;
				sboxes[876] <= 32'ha2ae0810;
				sboxes[877] <= 32'hdd6db224;
				sboxes[878] <= 32'h69852dfd;
				sboxes[879] <= 32'h9072166;
				sboxes[880] <= 32'hb39a460a;
				sboxes[881] <= 32'h6445c0dd;
				sboxes[882] <= 32'h586cdecf;
				sboxes[883] <= 32'h1c20c8ae;
				sboxes[884] <= 32'h5bbef7dd;
				sboxes[885] <= 32'h1b588d40;
				sboxes[886] <= 32'hccd2017f;
				sboxes[887] <= 32'h6bb4e3bb;
				sboxes[888] <= 32'hdda26a7e;
				sboxes[889] <= 32'h3a59ff45;
				sboxes[890] <= 32'h3e350a44;
				sboxes[891] <= 32'hbcb4cdd5;
				sboxes[892] <= 32'h72eacea8;
				sboxes[893] <= 32'hfa6484bb;
				sboxes[894] <= 32'h8d6612ae;
				sboxes[895] <= 32'hbf3c6f47;
				sboxes[896] <= 32'hd29be463;
				sboxes[897] <= 32'h542f5d9e;
				sboxes[898] <= 32'haec2771b;
				sboxes[899] <= 32'hf64e6370;
				sboxes[900] <= 32'h740e0d8d;
				sboxes[901] <= 32'he75b1357;
				sboxes[902] <= 32'hf8721671;
				sboxes[903] <= 32'haf537d5d;
				sboxes[904] <= 32'h4040cb08;
				sboxes[905] <= 32'h4eb4e2cc;
				sboxes[906] <= 32'h34d2466a;
				sboxes[907] <= 32'h115af84;
				sboxes[908] <= 32'he1b00428;
				sboxes[909] <= 32'h95983a1d;
				sboxes[910] <= 32'h6b89fb4;
				sboxes[911] <= 32'hce6ea048;
				sboxes[912] <= 32'h6f3f3b82;
				sboxes[913] <= 32'h3520ab82;
				sboxes[914] <= 32'h11a1d4b;
				sboxes[915] <= 32'h277227f8;
				sboxes[916] <= 32'h611560b1;
				sboxes[917] <= 32'he7933fdc;
				sboxes[918] <= 32'hbb3a792b;
				sboxes[919] <= 32'h344525bd;
				sboxes[920] <= 32'ha08839e1;
				sboxes[921] <= 32'h51ce794b;
				sboxes[922] <= 32'h2f32c9b7;
				sboxes[923] <= 32'ha01fbac9;
				sboxes[924] <= 32'he01cc87e;
				sboxes[925] <= 32'hbcc7d1f6;
				sboxes[926] <= 32'hcf0111c3;
				sboxes[927] <= 32'ha1e8aac7;
				sboxes[928] <= 32'h1a908749;
				sboxes[929] <= 32'hd44fbd9a;
				sboxes[930] <= 32'hd0dadecb;
				sboxes[931] <= 32'hd50ada38;
				sboxes[932] <= 32'h339c32a;
				sboxes[933] <= 32'hc6913667;
				sboxes[934] <= 32'h8df9317c;
				sboxes[935] <= 32'he0b12b4f;
				sboxes[936] <= 32'hf79e59b7;
				sboxes[937] <= 32'h43f5bb3a;
				sboxes[938] <= 32'hf2d519ff;
				sboxes[939] <= 32'h27d9459c;
				sboxes[940] <= 32'hbf97222c;
				sboxes[941] <= 32'h15e6fc2a;
				sboxes[942] <= 32'hf91fc71;
				sboxes[943] <= 32'h9b941525;
				sboxes[944] <= 32'hfae59361;
				sboxes[945] <= 32'hceb69ceb;
				sboxes[946] <= 32'hc2a86459;
				sboxes[947] <= 32'h12baa8d1;
				sboxes[948] <= 32'hb6c1075e;
				sboxes[949] <= 32'he3056a0c;
				sboxes[950] <= 32'h10d25065;
				sboxes[951] <= 32'hcb03a442;
				sboxes[952] <= 32'he0ec6e0e;
				sboxes[953] <= 32'h1698db3b;
				sboxes[954] <= 32'h4c98a0be;
				sboxes[955] <= 32'h3278e964;
				sboxes[956] <= 32'h9f1f9532;
				sboxes[957] <= 32'he0d392df;
				sboxes[958] <= 32'hd3a0342b;
				sboxes[959] <= 32'h8971f21e;
				sboxes[960] <= 32'h1b0a7441;
				sboxes[961] <= 32'h4ba3348c;
				sboxes[962] <= 32'hc5be7120;
				sboxes[963] <= 32'hc37632d8;
				sboxes[964] <= 32'hdf359f8d;
				sboxes[965] <= 32'h9b992f2e;
				sboxes[966] <= 32'he60b6f47;
				sboxes[967] <= 32'hfe3f11d;
				sboxes[968] <= 32'he54cda54;
				sboxes[969] <= 32'h1edad891;
				sboxes[970] <= 32'hce6279cf;
				sboxes[971] <= 32'hcd3e7e6f;
				sboxes[972] <= 32'h1618b166;
				sboxes[973] <= 32'hfd2c1d05;
				sboxes[974] <= 32'h848fd2c5;
				sboxes[975] <= 32'hf6fb2299;
				sboxes[976] <= 32'hf523f357;
				sboxes[977] <= 32'ha6327623;
				sboxes[978] <= 32'h93a83531;
				sboxes[979] <= 32'h56cccd02;
				sboxes[980] <= 32'hacf08162;
				sboxes[981] <= 32'h5a75ebb5;
				sboxes[982] <= 32'h6e163697;
				sboxes[983] <= 32'h88d273cc;
				sboxes[984] <= 32'hde966292;
				sboxes[985] <= 32'h81b949d0;
				sboxes[986] <= 32'h4c50901b;
				sboxes[987] <= 32'h71c65614;
				sboxes[988] <= 32'he6c6c7bd;
				sboxes[989] <= 32'h327a140a;
				sboxes[990] <= 32'h45e1d006;
				sboxes[991] <= 32'hc3f27b9a;
				sboxes[992] <= 32'hc9aa53fd;
				sboxes[993] <= 32'h62a80f00;
				sboxes[994] <= 32'hbb25bfe2;
				sboxes[995] <= 32'h35bdd2f6;
				sboxes[996] <= 32'h71126905;
				sboxes[997] <= 32'hb2040222;
				sboxes[998] <= 32'hb6cbcf7c;
				sboxes[999] <= 32'hcd769c2b;
				sboxes[1000] <= 32'h53113ec0;
				sboxes[1001] <= 32'h1640e3d3;
				sboxes[1002] <= 32'h38abbd60;
				sboxes[1003] <= 32'h2547adf0;
				sboxes[1004] <= 32'hba38209c;
				sboxes[1005] <= 32'hf746ce76;
				sboxes[1006] <= 32'h77afa1c5;
				sboxes[1007] <= 32'h20756060;
				sboxes[1008] <= 32'h85cbfe4e;
				sboxes[1009] <= 32'h8ae88dd8;
				sboxes[1010] <= 32'h7aaaf9b0;
				sboxes[1011] <= 32'h4cf9aa7e;
				sboxes[1012] <= 32'h1948c25c;
				sboxes[1013] <= 32'h2fb8a8c;
				sboxes[1014] <= 32'h1c36ae4;
				sboxes[1015] <= 32'hd6ebe1f9;
				sboxes[1016] <= 32'h90d4f869;
				sboxes[1017] <= 32'ha65cdea0;
				sboxes[1018] <= 32'h3f09252d;
				sboxes[1019] <= 32'hc208e69f;
				sboxes[1020] <= 32'hb74e6132;
				sboxes[1021] <= 32'hce77e25b;
				sboxes[1022] <= 32'h578fdfe3;
				sboxes[1023] <= 32'h3ac372e6;
				state <= NOP1;
			end
			else if(state == NOP1) begin
				n_ciclic_key <= n_key_r2;
				n_key_r <= 7'd0;
				state <= CYCLIC_KEY;
			end 
			else if( state == CYCLIC_KEY)begin
					if( n_ciclic_key < 7'd72)begin
						key_r[n_ciclic_key]  <=key_r[n_key_r];
						key_r2[n_ciclic_key] <=key_r[n_key_r];
						n_ciclic_key <= n_ciclic_key + 7'd1;
						if(n_key_r == (n_key_r2-7'd1))begin
							n_key_r <= 7'd0;
						end
						else begin
							n_key_r <= n_key_r + 7'd1;
						end
					end 					
					else begin
						n_key_r <= 7'd0;
						n_key_r2 <= 7'd72;
						count_XOR_PARRAY <= 5'd0;
						state <= INIT_XOR;
					end
				end
			else if(state == INIT_XOR) begin
				if(count_XOR_PARRAY < 5'd18)begin
					parray[count_XOR_PARRAY][31:24] <= parray[count_XOR_PARRAY][31:24]^ key_r[n_key_r];
					parray[count_XOR_PARRAY][23:16] <= parray[count_XOR_PARRAY][23:16]^ key_r[n_key_r+(7'd1)];
					parray[count_XOR_PARRAY][15:8]  <= parray[count_XOR_PARRAY][15:8] ^ key_r[n_key_r+(7'd2)];
					parray[count_XOR_PARRAY][ 7:0]  <= parray[count_XOR_PARRAY][ 7:0] ^ key_r[n_key_r+(7'd3)];
					count_XOR_PARRAY <= count_XOR_PARRAY + 5'd1;
					n_key_r <= n_key_r + 7'd4;
				end
				else begin
					n_key_r <= n_key_r2;
					count_XOR_PARRAY <= 5'd0;
					pApB <= 64'h0000000000000000;
					controlstate <= XOR_PARRAY_1;
					state <= CONTROL_EXPAND_KEY;
				end
			end
			else if(state == CONTROL_EXPAND_KEY) begin
				if(controlstate == XOR_PARRAY_1) begin
					if( S1S2 == 1'b0 )begin
						L <= salt_r[127:96]^pApB[63:32];//xL = xL ^ pArray[count_pArray];
						R <= salt_r[95:64]^pApB[31:0];
						S1S2 <= 1'b1;
					end
					else begin
						L 		<= salt_r[63:32] ^pApB[63:32];//xL = xL ^ pArray[count_pArray];
						R 		<= salt_r[31:0]^pApB[31:0];
						S1S2 	<= 1'b0;
					end
					count_round <= 4'd0;
					controlstate <= XOR_PARRAY_2;
					state <= PREF;
				end
				else if(controlstate == XOR_PARRAY_2)begin
					parray[count_XOR_PARRAY] <= pApB[63:32];
					parray[count_XOR_PARRAY+5'd1] <= pApB[31:0];
					if(count_XOR_PARRAY < 5'd16) begin
						count_XOR_PARRAY <= count_XOR_PARRAY+5'd2;
						controlstate <= XOR_PARRAY_1;
					end
					else begin
						count_XOR_SBOX <= 10'd0;
						controlstate <= XOR_SBOX_1;
					end
				end
				else if( controlstate == XOR_SBOX_1)begin
					if( S1S2 == 1'b0 )begin
						L 	<= salt_r[127:96]^pApB[63:32];//xL = xL ^ pArray[count_pArray];
						R 		<= salt_r[95:64]^pApB[31:0];
						S1S2 	<= 1'b1;
					end
					else begin
						L 	 <= salt_r[63:32]^pApB[63:32];//xL = xL ^ pArray[count_pArray];
						R 	 <= salt_r[31:0]^pApB[31:0];
						S1S2 <= 1'b0;
					end
					count_round <= 4'd0;
					controlstate <= XOR_SBOX_2;
					state <= PREF;
				end
				else 
				if(controlstate == XOR_SBOX_2)begin
					sboxes[count_XOR_SBOX] <= pApB[63:32];
					sboxes[count_XOR_SBOX+10'd1] <= pApB[31:0];
					if(count_XOR_SBOX == 10'd1022) begin
						salt_r <=128'h00000000000000000000000000000000;
						controlstate <= COST_1;
					end
					else begin
						count_XOR_SBOX <= count_XOR_SBOX+10'd2;
						controlstate <= XOR_SBOX_1;
					end
				end else 
				if(controlstate == CRYPT_1)begin
					ctext[2] <= pApB;
					if(cost_r> 64'd0)begin
						L 	 <= ctext[0][63:32];//xL = xL ^ pArray[count_pArray];
						R 	 <= ctext[0][31:0];
						count_round <= 4'd0;
						cost_r <= cost_r - 64'd1;
						controlstate <= CRYPT_2;
						state <= PREF;
					end else begin
						data_out[191:128] 	<= ctext[0];
						data_out[127:64] 	<= ctext[1];
						data_out[63:0] 		<= pApB;
						VALID_S <= 1'b1;
						state <= OUT_HANDSHAKE;
					end
				end else 
				if(controlstate == CRYPT_2)begin
						ctext[0] <= pApB;
						L 	 <= ctext[1][63:32];//xL = xL ^ pArray[count_pArray];
						R 	 <= ctext[1][31:0];
						count_round <= 4'd0;
						controlstate <= CRYPT_3;
						state <= PREF;
				end else 
				if(controlstate == CRYPT_3)begin
						ctext[1] <= pApB;
						L 	 <= ctext[2][63:32];//xL = xL ^ pArray[count_pArray];
						R 	 <= ctext[2][31:0];
						count_round <= 4'd0;
						controlstate <= CRYPT_1;
						state <= PREF;
				end else 
				if(controlstate ==  COST_1)begin
					if(cost_r==32'd0)begin
						ctext[0] <= 64'd5724761670021115458;
						ctext[1] <= 64'd7307212906740150867;
						ctext[2] <= 64'd7165923291934581364;
						pApB 	 <= 64'd7165923291934581364;
						cost_r   <= 32'd64;
						controlstate <= CRYPT_1;
					end else if(ctrl==1'b0)begin
						key_r[7'd0] <= salt_r2[127:120];
						key_r[7'd1] <= salt_r2[119:112];
						key_r[7'd2] <= salt_r2[111:104];
						key_r[7'd3] <= salt_r2[103:96];
						key_r[7'd4] <= salt_r2[95:88];
						key_r[7'd5] <= salt_r2[87:80];
						key_r[7'd6] <= salt_r2[79:72];
						key_r[7'd7] <= salt_r2[71:64];
						key_r[7'd8] <= salt_r2[63:56];
						key_r[7'd9] <= salt_r2[55:48];
						key_r[7'd10] <= salt_r2[47:40];
						key_r[7'd11] <= salt_r2[39:32];
						key_r[7'd12] <= salt_r2[31:24];
						key_r[7'd13] <= salt_r2[23:16];
						key_r[7'd14] <= salt_r2[15:8];
						key_r[7'd15] <= salt_r2[7:0];
						
						key_r[7'd16] <= salt_r2[127:120];
						key_r[7'd17] <= salt_r2[119:112];
						key_r[7'd18] <= salt_r2[111:104];
						key_r[7'd19] <= salt_r2[103:96];
						key_r[7'd20] <= salt_r2[95:88];
						key_r[7'd21] <= salt_r2[87:80];
						key_r[7'd22] <= salt_r2[79:72];
						key_r[7'd23] <= salt_r2[71:64];
						key_r[7'd24] <= salt_r2[63:56];
						key_r[7'd25] <= salt_r2[55:48];
						key_r[7'd26] <= salt_r2[47:40];
						key_r[7'd27] <= salt_r2[39:32];
						key_r[7'd28] <= salt_r2[31:24];
						key_r[7'd29] <= salt_r2[23:16];
						key_r[7'd30] <= salt_r2[15:8];
						key_r[7'd31] <= salt_r2[7:0];
						
						key_r[7'd32] <= salt_r2[127:120];
						key_r[7'd33] <= salt_r2[119:112];
						key_r[7'd34] <= salt_r2[111:104];
						key_r[7'd35] <= salt_r2[103:96];
						key_r[7'd36] <= salt_r2[95:88];
						key_r[7'd37] <= salt_r2[87:80];
						key_r[7'd38] <= salt_r2[79:72];
						key_r[7'd39] <= salt_r2[71:64];
						key_r[7'd40] <= salt_r2[63:56];
						key_r[7'd41] <= salt_r2[55:48];
						key_r[7'd42] <= salt_r2[47:40];
						key_r[7'd43] <= salt_r2[39:32];
						key_r[7'd44] <= salt_r2[31:24];
						key_r[7'd45] <= salt_r2[23:16];
						key_r[7'd46] <= salt_r2[15:8];
						key_r[7'd47] <= salt_r2[7:0];
						
						
						key_r[7'd48] <= salt_r2[127:120];
						key_r[7'd49] <= salt_r2[119:112];
						key_r[7'd50] <= salt_r2[111:104];
						key_r[7'd51] <= salt_r2[103:96];
						key_r[7'd52] <= salt_r2[95:88];
						key_r[7'd53] <= salt_r2[87:80];
						key_r[7'd54] <= salt_r2[79:72];
						key_r[7'd55] <= salt_r2[71:64];
						key_r[7'd56] <= salt_r2[63:56];
						key_r[7'd57] <= salt_r2[55:48];
						key_r[7'd58] <= salt_r2[47:40];
						key_r[7'd59] <= salt_r2[39:32];
						key_r[7'd60] <= salt_r2[31:24];
						key_r[7'd61] <= salt_r2[23:16];
						key_r[7'd62] <= salt_r2[15:8];
						key_r[7'd63] <= salt_r2[7:0];
						
						key_r[7'd64] <= salt_r2[127:120];
						key_r[7'd65] <= salt_r2[119:112];
						key_r[7'd66] <= salt_r2[111:104];
						key_r[7'd67] <= salt_r2[103:96];
						key_r[7'd68] <= salt_r2[95:88];
						key_r[7'd69] <= salt_r2[87:80];
						key_r[7'd70] <= salt_r2[79:72];
						key_r[7'd71] <= salt_r2[71:64];
						
						n_key_r <= 7'd0;
						ctrl <= 1'b1;
						S1S2 <= 1'b0;
						
						count_XOR_PARRAY <= 5'd0;
						pApB <= 64'h0000000000000000;
						
						controlstate <=  XOR_PARRAY_1;
						state <= INIT_XOR;
					end else begin
						key_r <= key_r2;
						n_key_r <= 7'd0;
						ctrl <= 1'b0;
						S1S2 <= 1'b0;
						cost_r <= cost_r - 32'd1;
						
						count_XOR_PARRAY <= 5'd0;
						pApB <= 64'h0000000000000000;
						controlstate <=  XOR_PARRAY_1;
						state <= INIT_XOR;
					end
				end
			end
			else if(state == PREF) begin 
				L <= L^parray[count_round];//xL = xL ^ pArray[count_round];
				state <= F1;
			end
			else if (state == F1)begin
				if( (sboxes[L[31:24]]+sboxes[10'd256+L[23:16]]) < 33'd4294967296)begin
					F_r <= (sboxes[L[31:24]]+sboxes[10'd256+L[23:16]]) ^ sboxes[10'd512+L[15:8]];
				end 
				else begin
					F_r <= (sboxes[L[31:24]]+sboxes[10'd256+L[23:16]] - 33'd4294967296)^ sboxes[10'd512+L[15:8]];
				end
				state <= F2;
			end
			else if (state == F2)begin
				if( (F_r + sboxes[10'd768+L[7:0]]) < 33'd4294967296)begin
					F_r <= (F_r + sboxes[10'd768+L[7:0]]);
				end 
				else begin
					F_r <= (F_r + sboxes[10'd768+L[7:0]]) - 33'd4294967296;
				end
				state <= XOR_XR_SWAP;
			end
			else if(state == XOR_XR_SWAP)begin
				L <= R^F_r;
				R <= L;
				if(count_round<4'd15)begin
					count_round <= count_round+4'd1;
					state <=PREF;
				end
				else begin
					state <= XOR_P17_P18;
				end
			end
			else if(state == XOR_P17_P18) begin
				pApB[63:32] <=  R ^  parray[5'd17];
				pApB[31:0]  <=  L ^  parray[5'd16];
				state <= CONTROL_EXPAND_KEY;
			end
			else if(state == OUT_HANDSHAKE)begin
				if(READY_S==1'b1) begin
					READY_E <= 1'b1;
					VALID_S <= 1'b0;
					state <= IN_HANDSHAKE;
				end
			end
		end	
		else begin
			/* Reset everything here */
			count_round <= 4'd0;
			state    <=   IN_HANDSHAKE;
			READY_E   <=  1'b1;
			VALID_S   <=  1'b0;
		end
	end
endmodule